import jinja2
import koji
import os
import yaml
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
from textwrap import dedent

from . import Modulemd

from .flatpak_builder import (FlatpakBuilder, FlatpakSourceInfo, FLATPAK_METADATA_ANNOTATIONS,
                              ModuleInfo)
from .installer import Installer
from .utils import check_call, die, log_call, warn, header, important, info, split_module_spec


class ContainerBuilder(object):
    def __init__(self, profile, containerspec, platform,
                 from_local=False, local_builds=[],
                 flatpak_metadata=FLATPAK_METADATA_ANNOTATIONS, install=False,
                 verbose=True):
        self.profile = profile
        self.containerspec = os.path.abspath(containerspec)
        self.platform = platform
        self.from_local = from_local
        self.local_builds = local_builds
        self.flatpak_metadata = flatpak_metadata
        self.install = install
        self.verbose = verbose

        with open(self.containerspec) as f:
            container_yaml = yaml.safe_load(f)

        compose_yaml = container_yaml.get('compose', {})
        modules = compose_yaml.get('modules', [])
        if not modules:
            die("No modules specified in the compose section of '{}'".format(self.containerspec))

        self.flatpak_yaml = container_yaml.get('flatpak', None)
        if not self.flatpak_yaml:
            die("No flatpak section in '{}'".format(containerspec))

        compose_yaml = container_yaml.get('compose', {})
        modules = compose_yaml.get('modules', [])
        if not modules:
            die("No modules specified in the compose section of '{}'".format(self.containerspec))

        if len(modules) > 1:
            warn("Multiple modules specified in compose section of '{}', using first".format(containerspec))

        self.module_spec = split_module_spec(modules[0])

    def build(self):
        workdir = tempfile.mkdtemp()
        try:
            dnf_conf = os.path.join(workdir, "dnf.conf")
            modules_file = os.path.join(workdir, "modules.yaml")

            module_build_id = self.module_spec.to_str(include_profile=False)

            env = dict(os.environ)
            if self.profile.mbs_config_file:
                env['MBS_CONFIG_FILE'] = self.profile.mbs_config_file
            if self.profile.mbs_config_section:
                env['MBS_CONFIG_SECTION'] = self.profile.mbs_config_section

            cmd = [
                'mbs-manager', 'run_with_compose',
                '--dnf-conf-output', dnf_conf,
                '--mmd-output', modules_file,
                '-m', module_build_id,
                '-m', 'platform:' + self.platform,
            ]
            for build in self.local_builds:
                cmd += ['-l', build]
            if self.from_local:
                cmd += ['-l', module_build_id]
            cmd += [
                '--', sys.argv[0], '-p', self.profile.name
            ]
            if self.verbose:
                cmd.append('-v')
            if self.verbose:
                cmd.append('-v')
            cmd += [
                'build-container-internal',
                '--flatpak-metadata', self.flatpak_metadata,
                '--containerspec', self.containerspec,
                '--platform', self.platform,
                '--dnf-conf', dnf_conf,
                '--modules-file', modules_file,
            ]
            if self.install:
                cmd.append('--install')

            return subprocess.call(cmd, env=env)
        finally:
            shutil.rmtree(workdir)

    def _get_better_mmd(self, mmd):
        name = mmd.get_module_name()
        stream = mmd.get_stream_name()
        version = mmd.get_version()
        context = mmd.get_context()

        koji_nvr = f"{name}-{stream.replace('-', '_')}-{version}.{context}"

        options = koji.read_config(profile_name=self.profile.koji_profile,
                                   user_config=self.profile.koji_config)
        session_opts = koji.grab_session_options(options)
        session = koji.ClientSession(options['server'], session_opts)

        build = session.getBuild(koji_nvr)
        if build:
            archives = session.listArchives(build['build_id'])
            for archive in archives:
                if archive['filename'] == 'modulemd.x86_64.txt':
                    import requests
                    pathinfo = koji.PathInfo(topdir=options['topurl'])
                    url = '/'.join((pathinfo.typedir(build, 'module'), archive['filename']))
                    response = requests.get(url)
                    response.raise_for_status()

                    return Modulemd.ModuleStream.read_string(response.text, True)

        return mmd

    def build_internal(self, dnf_conf, modules_file):
        header('BUILDING CONTAINER')
        important('container spec: {}'.format(self.containerspec))
        important('')

        mmd_index = Modulemd.ModuleIndex()
        mmd_index.update_from_file(modules_file, False)

        builds = {}

        mmds = mmd_index.search_streams()
        for mmd in mmds:
            mmd = self._get_better_mmd(mmd)

            rpms = [r + '.rpm' for r in mmd.get_rpm_artifacts()
                    if not koji.is_debuginfo(r.rsplit('-', 2)[0])]
            build = ModuleInfo(mmd.get_module_name(),
                               mmd.get_stream_name(),
                               mmd.get_version(),
                               mmd,
                               rpms)
            builds[build.name] = build
        base_build = builds[self.module_spec.name]

        builddir = os.path.expanduser("~/modulebuild/flatpaks")
        workdir = os.path.join(builddir, "{}-{}-{}".format(base_build.name, base_build.stream, base_build.version))
        if os.path.exists(workdir):
            info("Removing old output directory {}".format(workdir))
            shutil.rmtree(workdir)

        os.makedirs(workdir)
        info("Writing results to {}".format(workdir))

        source = FlatpakSourceInfo(self.flatpak_yaml, builds, base_build, self.module_spec.profile)

        builder = FlatpakBuilder(source, workdir, ".", flatpak_metadata=self.flatpak_metadata,
                                 verbose=self.verbose)

        component_label = self.flatpak_yaml.get('component', base_build.name)
        name_label = self.flatpak_yaml.get('name', base_build.name)
        version_label = base_build.stream
        release_label = base_build.version

        builder.add_labels({'name': name_label,
                            'com.redhat.component': component_label,
                            'version': version_label,
                            'release': release_label})

        env = jinja2.Environment(loader=jinja2.PackageLoader('flatpak_module_tools', 'templates'),
                                 autoescape=False)
        template = env.get_template('mock.cfg.j2')

        with open(dnf_conf) as f:
            dnf_conf_data = f.read().strip()

        dnf_conf_data = dnf_conf_data.replace('[main]',
                                              '[main]\nincludepkgs=' + ','.join(builder.get_includepkgs()))

        output_path = os.path.join(workdir, 'mock.cfg')

        template.stream(
            arch='x86_64',
            includepkgs=builder.get_includepkgs(),
            resultdir=workdir,
            dnf_conf=dnf_conf_data
        ).dump(output_path)

        finalize_script = dedent("""\
            #!/bin/sh
            set -ex
            userdel -f mockbuild
            groupdel mock
            """ + builder.get_cleanup_script()) + dedent("""\
            cd /
            exec tar cf - --anchored --exclude='./sys/*' --exclude='./proc/*' --exclude='./dev/*' --exclude='./run/*' ./
        """)

        finalize_script_path = os.path.join(workdir, 'finalize.sh')
        with open(finalize_script_path, 'w') as f:
            f.write(finalize_script)
            os.fchmod(f.fileno(), 0o0755)

        info('Initializing installation path')
        check_call(['mock', '-q', '-r', output_path, '--clean'],
                   log_filename=os.path.join(workdir, 'clean.log'))

        info('Installing packages')
        check_call(['mock', '-r', output_path, '--install'] + sorted(builder.get_install_packages()),
                   log_filename=os.path.join(workdir, 'mock-output.log'),
                   extra_log=os.path.join(workdir, 'root.log'))

        info('Cleaning and exporting filesystem')
        export_log = os.path.join(workdir, 'export.log')
        check_call(['mock', '-q', '-r', output_path, '--copyin', finalize_script_path, '/root/finalize.sh'],
                   log_filename=export_log)

        builder.root = "."

        args = ['mock', '-q', '-r', output_path, '--shell', '/root/finalize.sh']
        try:
            with open(export_log, 'ab') as export_log_stream:
                with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=export_log_stream) as process:
                    # When mock is using systemd-nspawn, systemd-nspawn dies with EPIPE if the output
                    # stream is closed before it exits, even if the child of systemd-nspawn isn't
                    # writing anything.
                    # https://github.com/systemd/systemd/issues/11533
                    filesystem_tar, manifestfile = builder._export_from_stream(process.stdout, close_stream=False)
        except tarfile.ReadError:
            # If the finalize script failed, then we expect reading from the tarfile to have failed,
            #   and we'll cleanly exit below.
            # if the finalize script succeeded, then we should raise an exception
            if process.returncode == 0:
                raise

        if process.returncode != 0:
            die("finalize.sh failed (exit status=%d)" % process.returncode)

        ref_name, outfile, tarred_outfile = builder.build_container(filesystem_tar)

        local_outname = "{}-{}-{}.oci.tar.gz".format(base_build.name, base_build.stream, base_build.version)

        info('Compressing result')
        with open(local_outname, 'wb') as f:
            subprocess.check_call(['gzip', '-c', tarred_outfile], stdout=f)

        important('Created ' + local_outname)

        if self.install:
            installer = Installer(self.profile)
            installer.set_source_path(local_outname)
            installer.install()

        return 0
