import click
import os
import re
import subprocess
import sys

from . import Modulemd

from .utils import die, warn, error, header, important, info

# ModuleBuilder is a straightforward, simple wrapper around:
#
#   mbs-manager build_module_locally

class ModuleBuilder(object):
    def __init__(self, profile, modulemd, stream, platform, local_builds=[], rebuild_strategy='only-changed'):
        self.profile = profile
        self.git_repo = self.find_git_repo()
        if self.git_repo:
            self.top_dir = self.git_repo
        else:
            self.top_dir = os.getcwd()

        if modulemd is not None:
            self.modulemd = modulemd
        else:
            self.modulemd = self.find_modulemd()

        if not self.modulemd.endswith(".yaml"):
            die("ModuleMD file '{}' does not have .yaml suffix")

        self.name = os.path.basename(self.modulemd)[:-5]
        if stream is not None:
            self.stream = stream
        else:
            self.stream = self.find_stream()

        self.platform = platform
        self.local_builds = local_builds
        self.rebuild_strategy = rebuild_strategy

    def find_git_repo(self):
        try:
            with open(os.devnull, 'w') as devnull:
                return subprocess.check_output(['git', 'rev-parse', '--show-toplevel'],
                                               stderr=devnull, encoding="UTF-8").strip()
        except subprocess.CalledProcessError:
            return None

    def find_modulemd(self):
        basename = os.path.basename(self.top_dir)
        modulemd = os.path.join(self.top_dir, basename + '.yaml')
        if not os.path.exists(modulemd):
            die("'{}' does not exist".format(modulemd))

        return modulemd

    def find_stream(self):
        mmd = Modulemd.ModuleStream.read_file(self.modulemd, True,
                                              None, None)
        mmd_stream = mmd.get_stream_name()

        # profile.allow_stream_override_from_scm means that when the user builds the
        # module in MBS, the stream in the modulemd file takes precedence over the
        # branch name. We need to behave the same to avoid confusion.
        if mmd_stream and (self.git_repo is None or self.profile.allow_stream_override_from_scm):
            return mmd_stream

        if self.git_repo is None:
            die("Not in a git repository, need to pass stream on the command line, "
                "or set it in the modulemd file")
        try:
            with open(os.devnull, 'w') as devnull:
                branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
                                                 stderr=devnull, encoding="UTF-8").strip()
        except subprocess.CalledProcessError:
            branch = None

        if branch is None or branch == 'HEAD':
            die("Cannot determine current git branch")

        if self.profile.allow_stream_override_from_scm:
            warn("No stream set in modulemd, using '%s'" % branch)
        elif mmd_stream and mmd_stream != branch:
            warn("Ignoring stream '%s' set in modulemd, using branch '%s' instead" %
                 (mmd_stream, branch))

        return branch

    def build(self):
        header('BUILDING MODULE')

        cmd = ['mbs-manager', 'build_module_locally',
               '--file', self.modulemd,
               '--stream', self.stream,
               '--rebuild-strategy', self.rebuild_strategy,
               '--buildrequires', 'platform:' + self.platform]
        for build_id in self.local_builds:
            cmd += ['--add-local-build', build_id]

        env = os.environ.copy()
        if self.profile.mbs_config_file is not None:
            env['MBS_CONFIG_FILE'] = self.profile.mbs_config_file
        if self.profile.mbs_config_section is not None:
            env['MBS_CONFIG_SECTION'] = self.profile.mbs_config_section
        process = subprocess.Popen(cmd, env=env)

        rv = process.wait()
        if rv != 0:
            error("mbs-manager build_module_locally failed")
            sys.exit(1)
